import string
import sys
import warnings
from contextlib import closing

import numpy as np
from gym import utils
from io import StringIO

from pogema.generator import generate_obstacles, generate_positions_and_targets_fast
from .grid_config import GridConfig


class Grid:

    def __init__(self, grid_config: GridConfig, add_artificial_border: bool = True, num_retries=10):

        self.config = grid_config

        if self.config.map is None:
            obstacles = generate_obstacles(self.config)
            starts_xy, finishes_xy = generate_positions_and_targets_fast(obstacles, self.config)
        else:
            obstacles = np.array([np.array(line) for line in self.config.map])
            starts_xy, finishes_xy = generate_positions_and_targets_fast(obstacles, self.config)

        if len(starts_xy) != len(finishes_xy):
            for attempt in range(num_retries):
                if len(starts_xy) == len(finishes_xy):
                    warnings.warn(f'Created valid configuration only with {attempt} attempts.', Warning, stacklevel=2)
                    break
                if self.config.map is None:
                    obstacles = generate_obstacles(self.config)
                starts_xy, finishes_xy = generate_positions_and_targets_fast(obstacles, self.config)

        # print(starts_xy, finishes_xy)
        # exit(0)
        if not starts_xy or not finishes_xy or len(starts_xy) != len(finishes_xy):
            raise OverflowError("Can't create task. Please check grid config, especially density, num_agent and map.")

        if add_artificial_border:
            r = self.config.obs_radius
            filled_obstacles = np.zeros(np.array(obstacles.shape) + r * 2)
            size, size = filled_obstacles.shape
            filled_obstacles[r - 1, r - 1:size - r + 1] = 1
            filled_obstacles[r - 1:size - r + 1, r - 1] = 1
            filled_obstacles[size - r, r - 1:size - r + 1] = 1
            filled_obstacles[r - 1:size - r + 1, size - r] = 1
            filled_obstacles[r:size - r, r:size - r] = obstacles

            obstacles = filled_obstacles

            starts_xy = [(x + r, y + r) for x, y in starts_xy]
            finishes_xy = [(x + r, y + r) for x, y in finishes_xy]

        filled_positions = np.zeros(obstacles.shape)
        for x, y in starts_xy:
            filled_positions[x, y] = 1

        self.obstacles = obstacles
        self.positions = filled_positions
        self.finishes_xy = finishes_xy
        self.positions_xy = starts_xy
        self.hidden_agents = set()

    def get_observation_shape(self):
        full_radius = self.config.obs_radius * 2 + 1
        return 2, full_radius, full_radius

    def get_num_actions(self):
        return len(self.config.MOVES)

    def get_obstacles(self, agent_id):
        x, y = self.positions_xy[agent_id]
        r = self.config.obs_radius
        return self.obstacles[x - r:x + r + 1, y - r:y + r + 1]

    def get_positions(self, agent_id):
        x, y = self.positions_xy[agent_id]
        r = self.config.obs_radius
        return self.positions[x - r:x + r + 1, y - r:y + r + 1]

    def get_target(self, agent_id):

        x, y = self.positions_xy[agent_id]
        fx, fy = self.finishes_xy[agent_id]
        if x == fx and y == fy:
            return 0.0, 0.0
        rx, ry = fx - x, fy - y
        dist = np.sqrt(rx ** 2 + ry ** 2)
        return rx / dist, ry / dist

    def get_square_target(self, agent_id):
        c = self.config
        full_size = self.config.obs_radius * 2 + 1
        result = np.zeros((full_size, full_size))
        x, y = self.positions_xy[agent_id]
        fx, fy = self.finishes_xy[agent_id]
        dx, dy = x - fx, y - fy

        dx = min(dx, c.obs_radius) if dx >= 0 else max(dx, -c.obs_radius)
        dy = min(dy, c.obs_radius) if dy >= 0 else max(dy, -c.obs_radius)
        result[c.obs_radius + dx, c.obs_radius + dy] = 1
        return result

    def render(self, mode='human'):
        outfile = StringIO() if mode == 'ansi' else sys.stdout
        chars = string.digits + string.ascii_letters + string.punctuation
        positions_map = {(x, y): id_ for id_, (x, y) in enumerate(self.positions_xy)}
        finishes_map = {(x, y): id_ for id_, (x, y) in enumerate(self.finishes_xy)}
        for line_index, line in enumerate(self.obstacles):
            out = ''
            for cell_index, cell in enumerate(line):
                if cell == self.config.FREE:
                    agent_id = positions_map.get((line_index, cell_index), None)
                    finish_id = finishes_map.get((line_index, cell_index), None)

                    if agent_id is not None:
                        out += str(utils.colorize(' ' + chars[agent_id % len(chars)] + ' ', color='gray', bold=True,
                                                  highlight=False))
                    elif finish_id is not None:
                        out += str(
                            utils.colorize('|' + chars[finish_id % len(chars)] + '|', 'white', highlight=False))
                    else:
                        out += str(utils.colorize(str(' . '), 'white', highlight=False))
                else:
                    out += str(utils.colorize(str('   '), 'cyan', bold=False, highlight=True))
            out += '\n'
            outfile.write(out)

        if mode != 'human':
            with closing(outfile):
                return outfile.getvalue()

    def move(self, agent_id, action):
        x, y = self.positions_xy[agent_id]

        self.positions[x, y] = self.config.FREE

        dx, dy = self.config.MOVES[action]
        if self.obstacles[x + dx, y + dy] == self.config.FREE and self.positions[x + dx, y + dy] == self.config.FREE:
            x += dx
            y += dy

        self.positions_xy[agent_id] = (x, y)
        self.positions[x, y] = self.config.OBSTACLE

    def on_goal(self, agent_id):
        return self.positions_xy[agent_id] == self.finishes_xy[agent_id]

    def hide_agent(self, agent_id):
        if agent_id in self.hidden_agents:
            return False

        self.hidden_agents.add(agent_id)

        x, y = self.positions_xy[agent_id]
        self.positions[x, y] = self.config.FREE

        return True


class CooperativeGrid(Grid):
    def move(self, agent_id, action):
        x, y = self.positions_xy[agent_id]
        fx, fy = self.finishes_xy[agent_id]
        dx, dy = self.config.MOVES[action]
        if self.obstacles[x + dx, y + dy] == self.config.FREE:
            if self.positions[x + dx, y + dy] == self.config.FREE:
                self.positions[x, y] = self.config.FREE
                x += dx
                y += dy
                self.positions[x, y] = self.config.OBSTACLE
        self.positions_xy[agent_id] = (x, y)
        return x == fx and y == fy



import json

import numpy as np
import gym
from gym.error import ResetNeeded
from gym.wrappers import TimeLimit

from pogema.animation import AnimationMonitor, MarlSvgVisualization
from pogema.grid import Grid, CooperativeGrid
from pogema.grid_config import GridConfig
from pogema.moving_obstacles import MovingObstaclesConfig, MovingObstaclesWrapper


class PogemaBase(gym.Env):

    def step(self, action):
        raise NotImplemented

    def reset(self):
        raise NotImplemented

    def __init__(self, config: GridConfig = GridConfig()):
        # noinspection PyTypeChecker
        self.grid: Grid = None
        self.config = config

        full_size = self.config.obs_radius * 2 + 1
        self.observation_space = gym.spaces.Box(0.0, 1.0, shape=(3, full_size, full_size))
        self.action_space = gym.spaces.Discrete(len(self.config.MOVES))

    def _get_agents_obs(self, agent_id=0):
        return np.concatenate([
            self.grid.get_obstacles(agent_id)[None],
            self.grid.get_positions(agent_id)[None],
            self.grid.get_square_target(agent_id)[None]
        ])

    def check_reset(self):
        if self.grid is None:
            raise ResetNeeded("Please reset environment first!")

    def render(self, mode='human'):
        self.check_reset()
        return self.grid.render(mode=mode)


class PogemaCoopFinish(PogemaBase):
    def __init__(self, config=GridConfig(num_agents=2)):
        super().__init__(config)
        # self.num_agents = self.config.num_agents
        self.is_multiagent = True

    def _obs(self):
        return [self._get_agents_obs(index) for index in range(self.config.num_agents)]

    def step(self, action: list):
        assert len(action) == self.config.num_agents
        rewards = []

        infos = [dict() for _ in range(self.config.num_agents)]

        dones = []
        for agent_idx in range(self.config.num_agents):
            agent_done = self.grid.move(agent_idx, action[agent_idx])

            if agent_done:
                rewards.append(1.0)
            else:
                rewards.append(0.0)

            dones.append(agent_done)

        obs = self._obs()
        return obs, rewards, dones, infos

    def reset(self):
        self.grid: CooperativeGrid = CooperativeGrid(grid_config=self.config)
        self.active = {agent_idx: True for agent_idx in range(self.config.num_agents)}
        return self._obs()


class Pogema(PogemaBase):
    def __init__(self, config=GridConfig(num_agents=2)):
        super().__init__(config)
        # self.num_agents = self.config.num_agents
        self.is_multiagent = True
        self.active = None

    def _obs(self):
        return [self._get_agents_obs(index) for index in range(self.config.num_agents)]

    def step(self, action: list):
        assert len(action) == self.config.num_agents
        rewards = []

        infos = [dict() for _ in range(self.config.num_agents)]

        dones = []
        for agent_idx in range(self.config.num_agents):
            if self.active[agent_idx]:
                self.grid.move(agent_idx, action[agent_idx])

            on_goal = self.grid.on_goal(agent_idx)
            if on_goal and self.active[agent_idx]:
                rewards.append(1.0)
            else:
                rewards.append(0.0)
            dones.append(on_goal)

        for agent_idx in range(self.config.num_agents):
            if self.grid.on_goal(agent_idx):
                self.grid.hide_agent(agent_idx)
                self.active[agent_idx] = False

            infos[agent_idx]['is_active'] = self.active[agent_idx]

        obs = self._obs()
        return obs, rewards, dones, infos

    def reset(self):
        self.grid: Grid = Grid(grid_config=self.config)
        self.active = {agent_idx: True for agent_idx in range(self.config.num_agents)}
        return self._obs()


class MultiTimeLimit(TimeLimit):
    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        self._elapsed_steps += 1
        if self._elapsed_steps >= self._max_episode_steps:
            for agent_idx in range(self.env.config.num_agents):
                info[agent_idx]["TimeLimit.truncated"] = not done[agent_idx]
                # log.debug('Set TimeLimit.truncated')
            done = [True] * self.num_agents
        return observation, reward, done, info


def main():
    grid_config = GridConfig(num_agents=40, size=8, obs_radius=5, density=0.0, seed=None,
                             map=[
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 0, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                                 [0, 0, 0, 1, 0, 0, 0],
                             ])
    # grid_config = GridConfig(num_agents=1, size=8, obs_radius=1, density=0.0, seed=None,)
    #
    # env = gym.make('Pogema-v0', config=grid_config)
    env = PogemaCoopFinish(config=grid_config)
    env = MultiTimeLimit(env, max_episode_steps=50)
    # env = AnimationMonitor(env)
    mo_config = MovingObstaclesConfig(
        num_obstacles=5,
        shake_r=1,
        size_range=[2, 4],
        density=0.7,
        hide_range=[8, 16],
        show_range=[8, 16],
    )
    # env = MovingObstaclesWrapper(env, mo_config)
    obs = env.reset()

    done = [False, ...]
    to_save = []
    for index in range(50):
        env.render()
        obs, reward, done, info = env.step([env.action_space.sample() for _ in range(env.config.num_agents)])
        to_save.append({
            'obstacles': list([list(map(int, line)) for line in env.grid.obstacles]),
            'positions_xy': list(env.grid.positions_xy),
            'finishes_xy': list(env.grid.finishes_xy),
            'obs_radius': env.config.obs_radius,
        })
        MarlSvgVisualization.draw_frame(env.grid.obstacles,
                                        env.grid.positions_xy,
                                        env.grid.finishes_xy, f'animations/{index}.svg',
                                        env.config.obs_radius)
        if all(done):
            break
    with open('out.json', "w") as f:
        json.dump(to_save, f)
    env.reset()


def m3():
    grid_config = GridConfig(num_agents=8, size=10, obs_radius=3, density=0.3, seed=42,)
    env = Pogema(grid_config, )
    env = MultiTimeLimit(env, max_episode_steps=100000)
    env.reset()

    done = [False, ...]
    to_save = []
    for index in range(50):
        env.render()
        obs, reward, done, info = env.step([env.action_space.sample() for _ in range(env.config.num_agents)])
        to_save.append({
            'obstacles': list([list(map(int, line)) for line in env.grid.obstacles]),
            'positions_xy': list(env.grid.positions_xy),
            'finishes_xy': list(env.grid.finishes_xy),
            'obs_radius': env.config.obs_radius,
        })
        MarlSvgVisualization.draw_frame(env.grid.obstacles,
                                        env.grid.positions_xy,
                                        env.grid.finishes_xy, f'animations/{index}.svg',
                                        env.config.obs_radius)
        if all(done):
            break


if __name__ == '__main__':
    m3()
    # main()



import gym

from pogema import GridConfig
from pogema.grid import Grid


class DenseRewardWrapper(gym.Wrapper):
    def __init__(self, env):
        super().__init__(env)
        self._best_distance = None

    def reset(self, *args, **kwargs):
        obs = self.env.reset(*args, **kwargs)
        config: GridConfig = self.env.config
        grid: Grid = self.env.grid
        self._best_distance = []
        for agent_idx in range(config.num_agents):
            x, y = grid.positions_xy[agent_idx]
            fx, fy = grid.finishes_xy[agent_idx]
            self._best_distance.append(self._manhattan_distance(x, y, fx, fy))
        return obs

    def step(self, action):
        obs, _, dones, infos = self.env.step(action)

        grid: Grid = self.env.grid
        rewards = []
        for agent_idx, best_distance in enumerate(self._best_distance):
            x, y = grid.positions_xy[agent_idx]
            fx, fy = grid.finishes_xy[agent_idx]
            distance = self._manhattan_distance(x, y, fx, fy)
            if distance < best_distance:
                self._best_distance[agent_idx] = distance
                rewards.append(1.0)
            else:
                rewards.append(0.0)

        return obs, rewards, dones, infos

    @staticmethod
    def _manhattan_distance(x1, y1, x2, y2):
        return abs(x1 - x2) + abs(y1 - y2)


class TrueRewardWrapper(gym.Wrapper):
    """
        Replaces reward function for each agent with 0 for non-terminal states and
        f* / (g + h) for terminal states, where
        f* - length of the optimal path for i-th agent (ignoring others),
        g - length of the agent path,
        h - length of the optimal path to the target from
            current position of the i-th agent
    """

    def __init__(self, env):
        super().__init__(env)
        self.shortest_paths = None
        self.step_cnt = None

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.step_cnt += 1

        grid: Grid = self.env.grid
        config: GridConfig = self.env.config

        reward = 0
        index = 0
        if done:
            x, y = grid.positions_xy[index]
            fx, fy = grid.finishes_xy[index]
            f_star = self.shortest_paths[index]
            g = self.step_cnt
            bfs_results = grid.obstacles.copy()
            bfs_results[bfs_results == 1] = -1

            self.bfs(bfs_results, config.MOVES, x, y, config.FREE)
            h = bfs_results[fx, fy] - 1
            reward = f_star / (g + h)

        return obs, reward, done, info

    @staticmethod
    def bfs(grid, moves, sx, sy, free_cell):
        q = [(sx, sy)]
        grid[sx, sy] = 1

        while len(q):
            x, y = q.pop(0)
            for dx, dy in moves:
                nx, ny = x + dx, y + dy
                if grid[nx, ny] == free_cell:
                    grid[nx, ny] = grid[x, y] + 1
                    q.append((nx, ny))

    def reset(self, **kwargs):
        self.step_cnt = 0
        result = self.env.reset(**kwargs)
        grid: Grid = self.env.grid
        config: GridConfig = self.env.config
        self.shortest_paths = []
        bfs_results = grid.obstacles.copy()
        bfs_results[bfs_results == 1] = -1
        for x, y in grid.positions_xy:
            bfs_results[x, y] = config.FREE

        for agent_index in range(config.num_agents):
            x, y = grid.positions_xy[agent_index]

            bfs_results[bfs_results > 0] = 0

            self.bfs(bfs_results, moves=config.MOVES, sx=x, sy=y, free_cell=config.FREE)
            self.shortest_paths.append(bfs_results[grid.finishes_xy[agent_index]] - 1)
        return result


class PrimalRewardWrapper(gym.Wrapper):

    def get_reward(self, agent_index, action):
        grid: Grid = self.env.grid

        x, y = grid.positions_xy[agent_index]
        fx, fy = grid.finishes_xy[agent_index]

        reward = -0.3
        if action == 0:
            reward = -0.5
        if x == fx and y == fy:
            reward = 20.0
        return reward

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        reward = self.get_reward(0, action)
        return obs, reward, done, info


class NormalizedPrimalRewardWrapper(PrimalRewardWrapper):
    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        reward = self.get_reward(0, action) / 20.0
        return obs, reward, done, info


class HammingRewardWrapper(gym.Wrapper):
    def __init__(self, env, max_rewarded_dist=32):
        super().__init__(env)
        self.max_rewarded_dist = max_rewarded_dist

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        agent_index = 0
        grid: Grid = self.env.grid

        if done:
            x, y = grid.positions_xy[agent_index]
            fx, fy = grid.finishes_xy[agent_index]
            hamming_dist = min(self.max_rewarded_dist, abs(x - fx) + abs(y - fy))
            reward = (self.max_rewarded_dist - hamming_dist) / self.max_rewarded_dist
        else:
            reward = 0

        return obs, reward, done, info



import gym


class SeedWrapper(gym.Wrapper):
    def __init__(self, env, seeds):
        self._seeds = seeds
        self._seed_cnt = 0
        super().__init__(env)

    def set_seed(self):
        self.env.config.seed = self._seeds[self._seed_cnt]
        self._seed_cnt = (self._seed_cnt + 1) % len(self._seeds)

    def reset(self, **kwargs):
        self.set_seed()
        return self.env.reset(**kwargs)



from gym import register
from pogema.grid_config import GridConfig

__version__ = '0.54'

__all__ = [
    'GridConfig',
]

register(
    id="Pogema-v0",
    entry_point="pogema.envs:Pogema",
)

register(
    id='PogemaCoopFinish-v0',
    entry_point="pogema.envs:PogemaCoopFinish"
)



from gym.wrappers import TimeLimit
from stable_baselines3.common.vec_env import SubprocVecEnv
# noinspection PyUnresolvedReferences
import pogema
import gym
from stable_baselines3 import PPO

import wandb

from pogema import GridConfig
from pogema.wrappers.reward_wrappers import TrueRewardWrapper, PrimalRewardWrapper
from pogema.wrappers.seed_wrapper import SeedWrapper


def make(config):
    env = gym.make("SingleAgentPogema-v0", config=config)
    env = TimeLimit(env, 16)
    return env


def run_sb3(config, use_wandb=False):
    policy_kwargs = dict(net_arch=[512, 512])
    project_name = 'pogema-sb3'
    if use_wandb:
        wb = wandb.init(project=project_name, sync_tensorboard=True, config=config)
    else:
        wb = None
    grid_config = GridConfig(num_agents=1, obs_radius=5, density=0.3, size=8)
    seeds = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    eval_env = SeedWrapper(gym.make("SingleAgentPogema-v0", config=grid_config), seeds=seeds)
    num_proc = 16

    if config['reward'] == 'true':
        env = SubprocVecEnv([lambda: TrueRewardWrapper(make(grid_config)) for _ in range(num_proc)])
    elif config['reward'] == 'success_rate':
        env = SubprocVecEnv([lambda: make(grid_config) for _ in range(num_proc)])
    elif config['reward'] == 'primal':
        env = SubprocVecEnv([lambda: PrimalRewardWrapper(make(grid_config)) for _ in range(num_proc)])
    else:
        raise KeyError

    agent = PPO('MlpPolicy', env, verbose=1, policy_kwargs=policy_kwargs,
                tensorboard_log='./pogema-sb3-2' + config['reward'])
    agent.learn(2000000, eval_env=eval_env, eval_freq=1000, n_eval_episodes=len(seeds))
    
    if wb:
        wb.finish()


def main():
    for reward in ['true', 'success_rate', 'primal']:
        config = {'reward': reward}
        run_sb3(config)


if __name__ == '__main__':
    main()



from queue import PriorityQueue

import gym
from pogema.animation import MarlSvgVisualization
from pogema.grid import Grid
from pogema.grid_config import GridConfig

import numpy as np
from pydantic import BaseModel


class MovingObstaclesConfig(BaseModel):
    density: float = 1.0
    shake_r: int = 0
    num_obstacles: int = 0
    show_range: list = [1, 1]
    hide_range: list = [1, 1]
    size_range: list = [2, 5]


class MovingObstacle:
    def __init__(self, x, y, size, rnd=None, mo_config=MovingObstaclesConfig()):
        if rnd is None:
            rnd = np.random.default_rng()
        self.rnd = rnd

        self.size = size
        self.mo_config: MovingObstaclesConfig = mo_config
        self.obstacle = self.rnd.binomial(1, self.mo_config.density, (self.size, self.size))

        self.x, self.y = x, y
        self.dx, self.dy = None, None
        self.shake_xy()

    def __lt__(self, other):
        return self.x < other.x and self.y < other.x

    def randint(self, low, high=None):
        if low == high:
            return low
        return self.rnd.integers(low=low, high=high)

    def get_xy(self):
        return self.x + self.dx, self.y + self.dy

    def shake_xy(self):
        r = self.mo_config.shake_r
        self.dx = self.randint(-r, r + 1)
        self.dy = self.randint(-r, r + 1)


class MovingObstaclesWrapper(gym.Wrapper):
    INF = 2 ** 20

    def __init__(self, env, mo_config=MovingObstaclesConfig(), rnd=None):
        super().__init__(env)

        if rnd is None:
            rnd = np.random.default_rng(self.env.config.seed)
        self.rnd = rnd

        self.mo_config: MovingObstaclesConfig = mo_config

        # noinspection PyTypeChecker
        self.hide_queue: PriorityQueue = None
        # noinspection PyTypeChecker
        self.show_queue: PriorityQueue = None
        self._t = None

        self._moving_obstacles = None

    def randint(self, low, high=None):
        if low == high:
            return low
        return self.rnd.integers(low=low, high=high)

    def reset(self):
        grid_config: GridConfig = self.env.config

        self.hide_queue: PriorityQueue = PriorityQueue()
        self.show_queue: PriorityQueue = PriorityQueue()
        self._t = 0

        observations = self.env.reset()
        grid: Grid = self.env.grid

        for _ in range(self.mo_config.num_obstacles):
            obstacle_size = self.randint(*self.mo_config.size_range)
            place_range = grid_config.obs_radius, grid_config.size + grid_config.obs_radius - obstacle_size + 1

            x, y = self.randint(*place_range), self.randint(*place_range)
            obstacle = MovingObstacle(x, y, obstacle_size, mo_config=self.mo_config, rnd=self.rnd)

            r = obstacle.size
            # lets make 100 attempts to place moving obstacle properly
            for _ in range(100):
                x, y = self.randint(*place_range), self.randint(*place_range)
                overlap = grid.obstacles[x:x + r, y:y + r].astype(int)
                overlap[overlap > 0] += 1
                intersection = overlap + obstacle.obstacle
                if np.count_nonzero(intersection == 1):
                    obstacle.x, obstacle.y = x, y
                    break

            self.show_queue.put((self.randint(*self.mo_config.show_range), obstacle))

        # create copy of the obstacles array with int type
        self._moving_obstacles = self.env.grid.obstacles.astype(int)
        self._moving_obstacles[self._moving_obstacles == grid_config.OBSTACLE] = self.INF
        return observations

    def update_overlap(self, x, y, r):
        g: Grid = self.grid
        overlap = self._moving_obstacles[x:x + r, y:y + r].copy()

        overlap[overlap > 0] = 1
        overlap[overlap <= 0] = 0

        # deal with obstacles on agents positions
        positions_overlap = g.positions[x:x + r, y:y + r].astype(int)
        overlap -= positions_overlap
        g.obstacles[x:x + r, y:y + r] = overlap.astype(float)

    def show_obstacle(self, o: MovingObstacle):
        o.shake_xy()
        x, y = o.get_xy()
        r = o.size
        self._moving_obstacles[x:x + r, y:y + r] += o.obstacle
        self.update_overlap(x, y, r)

    def hide_obstacle(self, o: MovingObstacle):
        x, y = o.get_xy()
        r = o.size
        self._moving_obstacles[x:x + r, y:y + r] -= o.obstacle
        self.update_overlap(x, y, r)

    @staticmethod
    def transfuse_queues(from_q, to_q, t, t_delta_func, process_func):

        while from_q.qsize() > 0:
            timer, obstacle = from_q.get()

            obstacle: MovingObstacle = obstacle
            if timer <= t:
                process_func(obstacle)
                new_time = t + max(1, t_delta_func())
                to_q.put((new_time, obstacle))
            else:
                from_q.put((timer, obstacle))
                return

    def step(self, actions):
        self._t += 1
        # show obstacles and schedule hiding
        self.transfuse_queues(from_q=self.show_queue, to_q=self.hide_queue, t=self._t,
                              t_delta_func=lambda: self.randint(*self.mo_config.show_range),
                              process_func=self.show_obstacle)

        # hide obstacles and schedule showing
        self.transfuse_queues(from_q=self.hide_queue, to_q=self.show_queue, t=self._t,
                              t_delta_func=lambda: self.randint(*self.mo_config.hide_range),
                              process_func=self.hide_obstacle)

        # self.fix_walls(self.env.grid.obstacles, self.env.config)

        observation, rewards, dones, infos = self.env.step(actions)

        # for x, y in self.env.grid.positions_xy:
        #     self.grid.obstacles[x, y] = self.env.config.FREE
        # for x, y in self.env.grid.finishes_xy:
        #     self.grid.obstacles[x, y] = self.env.config.FREE

        return observation, rewards, dones, infos


def main():
    grid_config = GridConfig(num_agents=1, size=32, obs_radius=5, density=0.0, seed=None)
    env = gym.make('Pogema-v0', config=grid_config)
    # env = AnimationMonitor(env)
    mo_config = MovingObstaclesConfig(
        num_obstacles=32,
        shake_r=5,
        size_range=(5, 10),
        density=0.7,
        hide_range=[8, 16],
        show_range=[8, 16],
    )
    env = MovingObstaclesWrapper(env, mo_config)
    obs = env.reset()

    done = [False, ...]
    for index in range(100):
        env.render()
        obs, reward, done, info = env.step([env.action_space.sample() for _ in range(env.config.num_agents)])
        MarlSvgVisualization.draw_frame(
            env.grid.obstacles,
            env.grid.positions_xy,
            env.grid.finishes_xy, f'animations/{index}.svg', env.config.obs_radius)
        if all(done):
            break
    env.reset()


if __name__ == '__main__':
    main()
